/*
 * libdevice-node
 *
 * Copyright (c) 2018 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include <libsyscommon/dbus-systemd.h>

#include <hw/board.h>
#include <hw/usb_gadget.h>

#ifndef EXPORT
#define EXPORT	__attribute__ ((visibility("default")))
#endif

static struct usb_function *_available_funcs[];

static void simple_cleanup_config(struct usb_configuration *config)
{
	if (!config)
		return;

	free(config->strs.config_str);

	if (config->funcs)
		free(config->funcs);

	free(config);
}

static void cleanup_gadget(struct usb_gadget *gadget)
{
	int i;

	if (!gadget)
		return;

	free(gadget->strs.manufacturer);
	free(gadget->strs.product);
	free(gadget->strs.serial);

	if (gadget->configs) {
		for (i = 0; gadget->configs[i]; ++i)
			simple_cleanup_config(gadget->configs[i]);

		free(gadget->configs);
	}

	free(gadget);
}

static int alloc_default_config(struct usb_configuration **_config)
{
	struct usb_configuration *config;

	config = calloc(1, sizeof(*config));
	if (!config)
		return -ENOMEM;

	config->attrs.bmAttributs = DEFAULT_BMATTRIBUTES;
	config->attrs.MaxPower = DEFAULT_MAX_POWER;

	/* TODO. Here is where to set the string used in config of configfs */

	*_config = config;

	return 0;
}

static int get_device_serial(char **out)
{
	struct hw_info *info;
	struct hw_common *common;
	struct hw_board *board;
	int ret;

	if (hw_get_info(BOARD_HARDWARE_DEVICE_ID, (const struct hw_info **)&info))
		return -ENOENT;

	ret = info->open(info, NULL, &common);
	if (ret < 0)
		return ret;

	board = container_of(common, struct hw_board, common);
	return board->get_device_serial(out);
}

static int alloc_default_gadget(struct usb_gadget **_gadget)
{
	struct usb_gadget *gadget;
	struct usb_configuration **configs;

	gadget = calloc(1, sizeof(*gadget));
	if (!gadget)
		goto out;

	gadget->attrs.idVendor = DEFAULT_VID;
	gadget->attrs.idProduct = DEFAULT_PID;
	gadget->attrs.bcdDevice = DEFAULT_BCD_DEVICE;

	gadget->strs.lang_code = DEFAULT_LANG;
	gadget->strs.manufacturer = strdup(DEFAULT_MANUFACTURER);
	gadget->strs.product = strdup(DEFAULT_PRODUCT);
	if (get_device_serial(&gadget->strs.serial) < 0)
		gadget->strs.serial = strdup(DEFAULT_SERIAL);

	if (!gadget->strs.manufacturer || !gadget->strs.product || !gadget->strs.serial)
		goto free_strs;

	/* slp-gadget use max 2 confiuration and NULL termination */
	configs = calloc(3, sizeof(*configs));
	if (!configs)
		goto free_strs;

	gadget->configs = configs;
	*_gadget = gadget;

	return 0;

free_strs:
	free(gadget->strs.manufacturer);
	free(gadget->strs.product);
	free(gadget->strs.serial);
	free(gadget);
out:
	return -ENOMEM;
}

static int id_to_gadget(struct usb_gadget_id *gadget_id, struct usb_gadget **_gadget)
{
	int ret;
	int i, j;
	int n_configs;
	struct usb_gadget *gadget;
	int functions[2][sizeof(gadget_id->function_mask)*8]; /* zero terminates */

	if (!gadget_id || !_gadget)
		return -EINVAL;

	ret = alloc_default_gadget(&gadget);
	if (ret)
		goto out;

	/*
	 * Currently all gadgets use inly single configuration but
	 * slp-gadget is capable to handle two of them
	 *
	 * Order of interfaces in configuration is significant
	 * so in this switch we sort our functions in a correct order
	 */
	switch (gadget_id->function_mask) {
	/* MTP, ACM, SDB */
	case USB_FUNCTION_MTP | USB_FUNCTION_ACM:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_MTP;
		functions[0][1] = USB_FUNCTION_ACM;
		functions[0][2] = 0;
		gadget->attrs.idProduct = 0x6860;
		break;

	case USB_FUNCTION_MTP | USB_FUNCTION_ACM | USB_FUNCTION_SDB:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_MTP;
		functions[0][1] = USB_FUNCTION_ACM;
		functions[0][2] = USB_FUNCTION_SDB;
		functions[0][3] = 0;
		gadget->attrs.idProduct = 0x6860;
		break;

	/* DIAG */
	case USB_FUNCTION_MTP | USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_DIAG:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_MTP;
		functions[0][1] = USB_FUNCTION_ACM;
		functions[0][2] = USB_FUNCTION_SDB;
		functions[0][3] = USB_FUNCTION_DIAG;
		functions[0][4] = 0;
		gadget->attrs.idProduct = 0x6860;
		break;

	/* RNDIS */
	case USB_FUNCTION_RNDIS:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_RNDIS;
		functions[0][1] = 0;
		gadget->attrs.idProduct = 0x6863;
		break;

	case USB_FUNCTION_RNDIS | USB_FUNCTION_DIAG:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_RNDIS;
		functions[0][1] = USB_FUNCTION_DIAG;
		functions[0][2] = 0;
		gadget->attrs.idProduct = 0x6864;
		break;

	case USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_RNDIS:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_RNDIS;
		functions[0][1] = USB_FUNCTION_ACM;
		functions[0][2] = USB_FUNCTION_SDB;
		functions[0][3] = 0;
		gadget->attrs.idProduct = 0x6864;
		break;

	/* RMNET */
	case USB_FUNCTION_DIAG | USB_FUNCTION_RMNET:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_DIAG;
		functions[0][1] = USB_FUNCTION_RMNET;
		functions[0][2] = 0;
		gadget->attrs.idProduct = 0x685d;
		break;

	/* DM */
	case USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_DM:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_ACM;
		functions[0][1] = USB_FUNCTION_SDB;
		functions[0][2] = USB_FUNCTION_DM;
		functions[0][3] = 0;
		gadget->attrs.idProduct = 0x6860;
		break;

	default:
		ret = -EINVAL;
		goto free_gadget;
	};

	for (j = 0; j < n_configs; ++j) {
		int n_funcs_in_config;
		struct usb_configuration *config;

		for (i = 0; functions[j][i]; ++i);
		n_funcs_in_config = i;

		ret = alloc_default_config(&config);
		if (ret)
			goto free_configs;

		gadget->configs[j] = config;
		config->funcs = calloc(n_funcs_in_config + 1, sizeof(void *));
		if (!config->funcs)
			goto free_configs;

		for (i = 0; functions[j][i]; ++i) {
			config->funcs[i] = find_usb_function_by_id(functions[j][i]);
			if (!config->funcs[i])
				goto free_configs;
		}
	}

	*_gadget = gadget;

	return 0;

free_configs:
free_gadget:
	cleanup_gadget(gadget);
out:
	return ret;
}

void rndis_handler(int enable)
{
	if (enable)
		(void)systemd_start_unit_wait_started("rndis.service", NULL, -1);
	else
		(void)systemd_stop_unit_wait_stopped("rndis.service", NULL, -1);
}

#define DEFINE_USB_FUNCTION(_id, _name, _is_functionfs, _service, _handler)  \
	static struct usb_function _##_name##_function = {                   \
		.id = _id,                                                   \
		.name = #_name,                                              \
		.instance = "default",                                       \
		.is_functionfs = _is_functionfs,                             \
		.service = _service,                                         \
		.handler = _handler,                                         \
	}

DEFINE_USB_FUNCTION(USB_FUNCTION_MTP,         mtp,         1, "mtp-responder", NULL);
DEFINE_USB_FUNCTION(USB_FUNCTION_ACM,         acm,         0, "data-router",   NULL);
DEFINE_USB_FUNCTION(USB_FUNCTION_SDB,         sdb,         1, "sdbd",          NULL);
DEFINE_USB_FUNCTION(USB_FUNCTION_RNDIS,       rndis,       0, "sshd",          rndis_handler);
DEFINE_USB_FUNCTION(USB_FUNCTION_DIAG,        diag,        0, NULL,            NULL);
DEFINE_USB_FUNCTION(USB_FUNCTION_CONN_GADGET, conn_gadget, 0, NULL,            NULL);
DEFINE_USB_FUNCTION(USB_FUNCTION_DM,          dm,          0, NULL,            NULL);
DEFINE_USB_FUNCTION(USB_FUNCTION_RMNET,       rmnet,       0, NULL,            NULL);

#undef DEFINE_USB_FUNCTION

/* Caution: index order of arrary is important, because simple_translator_open() uses it. */
static struct usb_function *_available_funcs[] = {
	[USB_FUNCTION_IDX_MTP]         = &_mtp_function,
	[USB_FUNCTION_IDX_ACM]         = &_acm_function,
	[USB_FUNCTION_IDX_SDB]         = &_sdb_function,
	[USB_FUNCTION_IDX_RNDIS]       = &_rndis_function,
	[USB_FUNCTION_IDX_DIAG]        = &_diag_function,
	[USB_FUNCTION_IDX_CONN_GADGET] = &_conn_gadget_function,
	[USB_FUNCTION_IDX_DM]          = &_dm_function,
	[USB_FUNCTION_IDX_RMNET]       = &_rmnet_function,
	[USB_FUNCTION_IDX_MAX]         = NULL /* An indicator to end the array */
};

struct usb_function *find_usb_function_by_id(int id)
{
	int i;

	for (i = 0; _available_funcs[i]; i++)
		if (_available_funcs[i]->id == id)
			return _available_funcs[i];

	return NULL;
}

struct usb_function *find_usb_function_by_name(const char *name)
{
	int i;

	if(!name || !name[0])
		return NULL;

	for (i = 0; _available_funcs[i]; i++)
		if (!strcmp(name, _available_funcs[i]->name))
			return _available_funcs[i];

	return NULL;
}

struct usb_function *find_usb_function_by_name_instance(const char *name, const char *instance)
{
	int i;

	if(!name || !name[0] || !instance || !instance[0])
		return NULL;

	for (i = 0; _available_funcs[i]; ++i)
		if (!strcmp(name, _available_funcs[i]->name) && !strcmp(instance, _available_funcs[i]->instance))
			return _available_funcs[i];

	return NULL;
}


EXPORT
int simple_translator_open(struct hw_info *info,
		const char *id, struct hw_common **common)
{
	struct usb_gadget_translator *simple_translator;

	if (!info || !common)
		return -EINVAL;

	simple_translator = calloc(1, sizeof(*simple_translator));
	if (!simple_translator)
		return -ENOMEM;

	simple_translator->common.info = info;
	simple_translator->id_to_gadget = id_to_gadget;
	simple_translator->cleanup_gadget = cleanup_gadget;

	/* Use mtp-responder-dummy.socket when there is no mtp-responser.socket.
	 *
	 * The mtp-responder.socket is special in the configfs environment.
	 * If mtp-responder.socket is missing, gadget configuration will fail.
	 * As a result, all usb operations do not work properly.
	 * So in environments that mtp doesn't support, use dummy mtp.
	 */
	if (access("/usr/lib/systemd/system/mtp-responder.socket", F_OK)) {
		_available_funcs[USB_FUNCTION_IDX_MTP]->service = "mtp-responder-dummy";
	}

	*common = &simple_translator->common;
	return 0;
}

EXPORT
int simple_translator_close(struct hw_common *common)
{
	struct usb_gadget_translator *simple_translator;

	if (!common)
		return -EINVAL;

	simple_translator = container_of(common, struct usb_gadget_translator,
					 common);

	free(simple_translator);
	return 0;
}
